﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ScatterSearch {
  /// <summary>
  /// An operator that updates the reference set.
  /// </summary>
  [Item("ReferenceSetUpdateMethod", "An operator that updates the reference set.")]
  [StorableClass]
  public sealed class ReferenceSetUpdateMethod : SingleSuccessorOperator {
    #region Parameter properties
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public IValueLookupParameter<IntValue> NumberOfHighQualitySolutionsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["NumberOfHighQualitySolutions"]; }
    }
    public IValueLookupParameter<IntValue> ReferenceSetSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["ReferenceSetSize"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    public IValueLookupParameter<DiversityCalculator> DiversityCalculatorParameter {
      get { return (IValueLookupParameter<DiversityCalculator>)Parameters["DiversityCalculator"]; }
    }
    #endregion

    #region Properties
    private IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    private IntValue NumberOfHighQualitySolutions {
      get { return NumberOfHighQualitySolutionsParameter.ActualValue; }
    }
    private IntValue ReferenceSetSize {
      get { return ReferenceSetSizeParameter.ActualValue; }
    }
    private IItem Target {
      get { return TargetParameter.ActualValue; }
    }
    private DiversityCalculator DiversityCalculator {
      get { return DiversityCalculatorParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private ReferenceSetUpdateMethod(bool deserializing) : base(deserializing) { }
    private ReferenceSetUpdateMethod(ReferenceSetUpdateMethod original, Cloner cloner) : base(original, cloner) { }
    public ReferenceSetUpdateMethod() : base() { Initialize(); }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ReferenceSetUpdateMethod(this, cloner);
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfHighQualitySolutions"));
      Parameters.Add(new ValueLookupParameter<IntValue>("ReferenceSetSize"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      Parameters.Add(new ValueLookupParameter<DiversityCalculator>("DiversityCalculator"));
      #endregion
      TargetParameter.ActualName = "KnapsackSolution"; // temporary solution for the knapsack problem
    }

    public override IOperation Apply() {
      var population = new Dictionary<IScope, double>();
      foreach (var pScope in CurrentScope.SubScopes[0].SubScopes) {
        int diversity = 0;
        var pSol = pScope.Variables[TargetParameter.ActualName].Value;
        foreach (var rScope in CurrentScope.SubScopes[1].SubScopes) {
          var rSol = rScope.Variables[TargetParameter.ActualName].Value;
          diversity += DiversityCalculator.ExecuteCalculation(pSol, rSol);
        }
        population[pScope] = diversity;
      }
      foreach (var entry in population.OrderByDescending(x => x.Value).Take(ReferenceSetSize.Value - NumberOfHighQualitySolutions.Value)) {
        CurrentScope.SubScopes[1].SubScopes.Add(entry.Key);
        CurrentScope.SubScopes[0].SubScopes.Remove(entry.Key);
      }
      return base.Apply();
    }
  }
}
